package gpl.pierrick.brihaye.aramorph;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

/**
 * This class contains a main method that transforms dictionaries and
 * compatability tables from text format to XML format.
 * @author Ahmed Saad
 */
public class DictTables2XML_SAX {
	/** A pattern which is used in parsing the dictionary file*/
	private static Pattern p = Pattern.compile(".*" + "<pos>(.+?)</pos>" + ".*");
	/**This map is used for translation of dictionaries*/
	private static Map translationMap = getTranslationSet();
	/**Encoding of output XML*/
	private static final String XML_ENCODING = "UTF-8";

	/*Some constants used for writing the XML file*/
	private static final String ROOT = "root";
	private static final String VALUE = "value";
	private static final String ENTRY = "entry";
	private static final String LEMMA = "lemma";
	private static final String ID = "id";
	private static final String UNVOCALIZED = "unvocalized";
	private static final String VOCALIZED = "vocalized";
	private static final String MORPHOLOGICAL_CATEGORY = "morphological-category";
	private static final String GLOSSES = "glosses";
	private static final String GLOSS = "gloss";
	private static final String GRAMMATICAL_CATEGORIES = "grammatical-categories";
	private static final String GRAMMATICAL_CATEGORY = "grammatical-category";
	private static final String LEMMA_ID = "lemmaID";
	

	/**
	 * An entry point to the converter.
	 * Automatically converts all dictionary to XML.
	 * @param args
	 */
	public static void main(String[] args) {
		new DictTables2XML_SAX();
	}
	
	/**
	 * Default constructor, converts all dictionaries to XML.
	 */
	public DictTables2XML_SAX() {
		convertAll();
	}
	
	/**
	 * Converts all dictionaries to XML
	 */
	public void convertAll() {
		System.out.println("Transforming dictionaries and compatability tables to XML...");
		// transform 3 lexicons
		transformDictionary("dictPrefixes", this.getClass().getResourceAsStream("dictionaries/dictPrefixes"));
		transformDictionary("dictStems", this.getClass().getResourceAsStream("dictionaries/dictStems"));
		transformDictionary("dictSuffixes", this.getClass().getResourceAsStream("dictionaries/dictSuffixes"));
		//transform 3 compatibility tables
		transformCompatibilityTable("prefix","stem","tableAB",this.getClass().getResourceAsStream("dictionaries/tableAB"));
		transformCompatibilityTable("prefix","suffix","tableAC", this.getClass().getResourceAsStream("dictionaries/tableAC"));
		transformCompatibilityTable("stem","suffix","tableBC", this.getClass().getResourceAsStream("dictionaries/tableBC"));		
		//write status message
		System.out.println("... done.");
	}
	
	/**
	 * Returns the translation set used to translated glosses.
	 * @return a map that maps each character to a character or string
	 */
	private static Map getTranslationSet() {
		Map s = new HashMap(61);
		
		s.put(new Character(';'),new Character('/')); //TODO : is it necessary ?
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('A'));
		s.put(new Character(''),new Character('C'));
		s.put(new Character(''),new Character('E'));
		s.put(new Character(''),new Character('E'));
		s.put(new Character(''),new Character('E'));
		s.put(new Character(''),new Character('E'));
		s.put(new Character(''),new Character('I'));
		s.put(new Character(''),new Character('I'));
		s.put(new Character(''),new Character('I'));
		s.put(new Character(''),new Character('I'));
		s.put(new Character(''),new Character('N'));
		s.put(new Character(''),new Character('O'));
		s.put(new Character(''),new Character('O'));
		s.put(new Character(''),new Character('O'));
		s.put(new Character(''),new Character('O'));
		s.put(new Character(''),new Character('O'));
		s.put(new Character(''),new Character('U'));
		s.put(new Character(''),new Character('U'));
		s.put(new Character(''),new Character('U'));
		s.put(new Character(''),new Character('U'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('a'));
		s.put(new Character(''),new Character('c'));
		s.put(new Character(''),new Character('e'));
		s.put(new Character(''),new Character('e'));
		s.put(new Character(''),new Character('e'));
		s.put(new Character(''),new Character('e'));
		s.put(new Character(''),new Character('i'));
		s.put(new Character(''),new Character('i'));
		s.put(new Character(''),new Character('i'));
		s.put(new Character(''),new Character('i'));
		s.put(new Character(''),new Character('n'));
		s.put(new Character(''),new Character('o'));
		s.put(new Character(''),new Character('o'));
		s.put(new Character(''),new Character('o'));
		s.put(new Character(''),new Character('o'));
		s.put(new Character(''),new Character('o'));
		s.put(new Character(''),new Character('u'));
		s.put(new Character(''),new Character('u'));
		s.put(new Character(''),new Character('u'));
		s.put(new Character(''),new Character('u'));
		s.put(new Character(''),new String("AE"));
		s.put(new Character(''),new String("Sh"));
		s.put(new Character(''),new String("Zh"));
		s.put(new Character(''),new String("ss"));
		s.put(new Character(''),new String("ae"));
		s.put(new Character(''),new String("sh"));
		s.put(new Character(''),new String("zh"));
		return s;
	}

	/**
	 * Transforms a dictionary written as text file to XML file.
	 * The output file differs for prefix, suffix and stem.
	 * For prefix and suffix, the output file is a list of &lt;entry&gt; tags.
	 * For stems, it is hierarchy of &lt;root&gt; tags, each containing  &lt;lemma&gt; tags.
	 *  &lt;root&gt; tags contain  &lt;entry&gt; tags.
	 * @param name the name of the dictionary. It is used to write the ouput file
	 * @param is InputStream for the text file.
	 * @throws RuntimeException when an IO error occurs reading from InputStream
	 * @see #transformDictionary2(String, InputStream)
	 */
	private void transformDictionary(String name, InputStream is) throws RuntimeException{ //TODO : should be static
		boolean inLemma = false;
		boolean inRoot = false;
		try {
			//prepare output file
			StreamResult out = new StreamResult(name+".xml");
			//prpare XML parser
			SAXTransformerFactory tf = (SAXTransformerFactory) SAXTransformerFactory.newInstance();
			TransformerHandler th = tf.newTransformerHandler();
			Transformer serializer = th.getTransformer();
			serializer.setOutputProperty(OutputKeys.ENCODING,XML_ENCODING);
			//pretty XML output
		    serializer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
		    serializer.setOutputProperty(OutputKeys.INDENT,"yes");
		    //attach output file to XML
		    th.setResult(out);
		    //initialize attributes
		    AttributesImpl atts = new AttributesImpl();

		    //start writing the output file
			// Create an XML Document
			th.startDocument();
			th.startElement("","","dictionary",atts);
			
			//HashSet lemmas = new HashSet();
			int forms = 0;
			String lemmaID = "";
			System.out.print("Transforming dictionary : " + name + " ");
			
			LineNumberReader IN = new LineNumberReader(new InputStreamReader(is,"ISO8859_1"));
			String line = null;
			while ((line = IN.readLine()) != null) {
				if ((IN.getLineNumber() % 1000) == 1) System.out.print(".");
				// new lemma
				if (line.startsWith(";; ")) {
					//append previous lemma
					if (inLemma)
						th.endElement("","",LEMMA);
					//remove the beginning ;;
					lemmaID = line.substring(3);
					//arabize lemmaId
					lemmaID = AraMorph.arabizeWord(lemmaID);
					//write to XML
					atts.addAttribute("","",ID,"CDATA",lemmaID);
					th.startElement("","",LEMMA,atts);
					atts.clear();
					//mark current state as inLemma, for this tag to be closed later
					inLemma = true;
				}
				// root
				else if (line.startsWith(";--- ")) {
					//append previous root
					if (inRoot) {
						if (inLemma)
							th.endElement("","",LEMMA);
						inLemma = false;
						lemmaID = "";
						th.endElement("","",ROOT);
					}
					//remove the beginning ;---
					String root = line.substring(5).trim();
					//remove any string after the root "see also ... etc."
					root = root.split(" ",2)[0].trim();
					//translate to arabic
					root = AraMorph.arabizeWord(root);
					//write to XML
					atts.addAttribute("","",VALUE,"CDATA",root);
					th.startElement("","",ROOT,atts);
					atts.clear();
					//mark current state as inRoot, for this tag to be closed later
					inRoot = true;
				}
				//comment
				else if (line.startsWith(";")) {
					String comment = line.substring(1).trim();
					th.comment(comment.toCharArray(),0,comment.length());
				}
				else {
					String split[] = line.split("\t",-1); //-1 to avoid triming of trail values
					
					//a little error-checking won't hurt :
					if (split.length != 4) {
						throw new RuntimeException("Entry in " + name + " (line " + IN.getLineNumber() + ") doesn't have 4 fields (3 tabs)");
					}
					String entry = AraMorph.arabizeWord(split[0]); // get the entry for use as key
					String vocalization = AraMorph.arabizeWord(split[1]);
					String morphology = split[2];
					String glossPOS = split[3];
					
					String gloss;
					String POS;
					
					//Pattern p;
					Matcher m;
					
					// two ways to get the POS info:
					// (1) explicitly, by extracting it from the gloss field:
					//p = Pattern.compile(".*" + "<pos>(.+?)</pos>" + ".*");
					m = p.matcher(glossPOS);
					if (m.matches()) {
						POS = m.group(1); //extract POS from glossPOS
						gloss = glossPOS; //we clean up the gloss later (see below)
					}
					// (2) by deduction: use the morphology (and sometimes the voc and gloss) to deduce the appropriate POS
					else {
						// we need the gloss to guess proper names
						gloss = glossPOS; 
						// null prefix or suffix
						if (morphology.matches("^(Pref-0|Suff-0)$")) { 
							POS = "";
						}
						else if (morphology.matches("^F" + ".*")) {
							POS = vocalization + "/FUNC_WORD";
						}
						else if (morphology.matches("^IV" + ".*")) {
							POS = vocalization + "/VERB_IMPERFECT";
						}
						else if (morphology.matches("^PV" + ".*")) {
							POS = vocalization + "/VERB_PERFECT";
						}
						else if (morphology.matches("^CV" + ".*")) {
							POS = vocalization + "/VERB_IMPERATIVE";
						}						
						else if (morphology.matches("^N" + ".*")) {
							// educated guess (99% correct)
							if (gloss.matches("^[A-Z]" + ".*")) {
								POS = vocalization + "/NOUN_PROP";
							}
							// (was NOUN_ADJ: some of these are really ADJ's and need to be tagged manually)
							else if (vocalization.matches(".*" + "iy~$")) { 
								POS = vocalization + "/NOUN";
							}
							else 
								POS = vocalization + "/NOUN";
						}
						else {
							throw new RuntimeException("No POS can be deduced in " + name + " (line " + IN.getLineNumber() + ")");
						}
					}
					
					// clean up the gloss: remove POS info and extra space, and convert upper-ASCII  to lower (it doesn't convert well to UTF-8)
					gloss = gloss.replaceFirst("<pos>.+?</pos>","");
					gloss = gloss.trim();
					
					gloss = translate(gloss);

					// note that although we read 4 fields from the dict we now save 5 fields in the hash table
					// because the info in last field, glossPOS, was split into two: gloss and POS
					DictionaryEntry de = new DictionaryEntry(entry, lemmaID, vocalization, morphology, gloss, POS);

					//add dictionary entry to xml file
					th.startElement("","",ENTRY,atts);
					//add unvocalized
					th.startElement("","",UNVOCALIZED,atts);
					th.characters(de.getEntry().toCharArray(),0,de.getEntry().length());
					th.endElement("","",UNVOCALIZED);
					//add vocalized
					th.startElement("","",VOCALIZED,atts);
					th.characters(de.getVocalization().toCharArray(),0,de.getVocalization().length());
					th.endElement("","",VOCALIZED);
					//add morphological category
					th.startElement("","",MORPHOLOGICAL_CATEGORY,atts);
					th.characters(de.getMorphology().toCharArray(),0,de.getMorphology().length());
					th.endElement("","",MORPHOLOGICAL_CATEGORY);
					//add glosses
					th.startElement("","",GLOSSES,atts);
					String[] glo = de.getGlosses();
					for (int i = 0; i < glo.length;i++) {
						th.startElement("","",GLOSS,atts);
						th.characters(glo[i].toCharArray(),0,glo[i].length());
						th.endElement("","",GLOSS);
					}
					th.endElement("","",GLOSSES);
					//add grammatical categories
					th.startElement("","",GRAMMATICAL_CATEGORIES,atts);
					String[] pos = de.getPOS();
					for (int i = 0; i < pos.length;i++) {
						th.startElement("","",GRAMMATICAL_CATEGORY,atts);
						String[] parts = pos[i].split("/");
						parts[0] = AraMorph.arabizeWord(parts[0]);
						String arabizedPOS = parts[0];
						if (parts.length > 1)
							arabizedPOS += "/"+parts[1];
						th.characters(arabizedPOS.toCharArray(),0,arabizedPOS.length());
						th.endElement("","",GRAMMATICAL_CATEGORY);
					}
					th.endElement("","",GRAMMATICAL_CATEGORIES);
					//append dictionary entry to xml file
					th.endElement("","",ENTRY);
					
					if (!inLemma)
						System.err.println("Found an entry without lemma in line: "+IN.getLineNumber());
					forms++;
				}
			}
			IN.close();
			//finalize xml file
			//append previous root
			if (inRoot) {
				if (inLemma)
					th.endElement("","",LEMMA);
				inLemma = false;
				th.endElement("","",ROOT);
			}
			th.endElement("","","dictionary");
			th.endDocument();
			//write status message
			System.out.println();
			//if (!"".equals(lemmaID)) System.out.print(lemmas.size() + " lemmas and ");
			//System.out.println(xmlRoot.getChildNodes().getLength() + " entries totalizing " + forms + " forms");
		} catch (TransformerConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			throw new RuntimeException("Can not open : " + name);
		}
	}

	/**
	 * Very similar to {@link #transformDictionary(String, InputStream)}, but writes Stems dictionary
	 * as a list of &lt;entry&gt; tags instead of a hierarchy structure.
	 * @param name the name of the dictionary. It is used to write the ouput file
	 * @param is InputStream for the text file.
	 * @throws RuntimeException when an IO error occurs reading from InputStream
	 * @see #transformDictionary(String, InputStream)
	 */
	private void transformDictionary2(String name, InputStream is) throws RuntimeException { //TODO : should be static
		try {
			//prepare output file
			StreamResult out = new StreamResult(name+".xml");
			//prpare XML parser
			SAXTransformerFactory tf = (SAXTransformerFactory) SAXTransformerFactory.newInstance();
			TransformerHandler th = tf.newTransformerHandler();
			Transformer serializer = th.getTransformer();
			serializer.setOutputProperty(OutputKeys.ENCODING,XML_ENCODING);
			//pretty XML output
		    serializer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
		    serializer.setOutputProperty(OutputKeys.INDENT,"yes");
		    //attach output file to XML
		    th.setResult(out);
		    //initialize attributes
		    AttributesImpl atts = new AttributesImpl();

		    //start writing the output file
			// Create an XML Document
			th.startDocument();
			th.startElement("","","dictionary",atts);
			
			//HashSet lemmas = new HashSet();
			int forms = 0;
			String lemmaID = "";
			String root = "";
			System.out.print("Transforming dictionary : " + name + " ");
			
			LineNumberReader IN = new LineNumberReader(new InputStreamReader(is,"ISO8859_1"));
			String line = null;
			while ((line = IN.readLine()) != null) {
				if ((IN.getLineNumber() % 1000) == 1) System.out.print(".");
				// new lemma
				if (line.startsWith(";; ")) {
					//remove the beginning ;;
					lemmaID = line.substring(3);
					//arabize lemmaId
					lemmaID = AraMorph.arabizeWord(lemmaID);
				}
				// root
				else if (line.startsWith(";--- ")) {
					//remove the beginning ;---
					root = line.substring(5).trim();
					//remove any string after the root "see also ... etc."
					root = root.split(" ",2)[0].trim();
					//translate to arabic
					root = AraMorph.arabizeWord(root);
					//reset lemma
					lemmaID = "";
				}
				//comment
				else if (line.startsWith(";")) {
					String comment = line.substring(1).trim();
					th.comment(comment.toCharArray(),0,comment.length());
				} else {
					String split[] = line.split("\t",-1); //-1 to avoid triming of trail values
					
					//a little error-checking won't hurt :
					if (split.length != 4) {
						throw new RuntimeException("Entry in " + name + " (line " + IN.getLineNumber() + ") doesn't have 4 fields (3 tabs)");
					}
					String entry = AraMorph.arabizeWord(split[0]); // get the entry for use as key
					String vocalization = AraMorph.arabizeWord(split[1]);
					String morphology = split[2];
					String glossPOS = split[3];
					
					String gloss;
					String POS;
					
					//Pattern p;
					Matcher m;
					
					// two ways to get the POS info:
					// (1) explicitly, by extracting it from the gloss field:
					//p = Pattern.compile(".*" + "<pos>(.+?)</pos>" + ".*");
					m = p.matcher(glossPOS);
					if (m.matches()) {
						POS = m.group(1); //extract POS from glossPOS
						gloss = glossPOS; //we clean up the gloss later (see below)
					}
					// (2) by deduction: use the morphology (and sometimes the voc and gloss) to deduce the appropriate POS
					else {
						// we need the gloss to guess proper names
						gloss = glossPOS; 
						// null prefix or suffix
						if (morphology.matches("^(Pref-0|Suff-0)$")) { 
							POS = "";
						}
						else if (morphology.matches("^F" + ".*")) {
							POS = vocalization + "/FUNC_WORD";
						}
						else if (morphology.matches("^IV" + ".*")) {
							POS = vocalization + "/VERB_IMPERFECT";
						}
						else if (morphology.matches("^PV" + ".*")) {
							POS = vocalization + "/VERB_PERFECT";
						}
						else if (morphology.matches("^CV" + ".*")) {
							POS = vocalization + "/VERB_IMPERATIVE";
						}						
						else if (morphology.matches("^N" + ".*")) {
							// educated guess (99% correct)
							if (gloss.matches("^[A-Z]" + ".*")) {
								POS = vocalization + "/NOUN_PROP";
							}
							// (was NOUN_ADJ: some of these are really ADJ's and need to be tagged manually)
							else if (vocalization.matches(".*" + "iy~$")) { 
								POS = vocalization + "/NOUN";
							}
							else 
								POS = vocalization + "/NOUN";
						}
						else {
							throw new RuntimeException("No POS can be deduced in " + name + " (line " + IN.getLineNumber() + ")");
						}
					}
					
					// clean up the gloss: remove POS info and extra space, and convert upper-ASCII  to lower (it doesn't convert well to UTF-8)
					gloss = gloss.replaceFirst("<pos>.+?</pos>","");
					gloss = gloss.trim();
					
					gloss = translate(gloss);

					// note that although we read 4 fields from the dict we now save 5 fields in the hash table
					// because the info in last field, glossPOS, was split into two: gloss and POS
					DictionaryEntry de = new DictionaryEntry(entry, lemmaID, vocalization, morphology, gloss, POS);

					//add dictionary entry to xml file
					th.startElement("","",ENTRY,atts);
					//add unvocalized
					th.startElement("","",UNVOCALIZED,atts);
					th.characters(de.getEntry().toCharArray(),0,de.getEntry().length());
					th.endElement("","",UNVOCALIZED);
					//add vocalized
					th.startElement("","",VOCALIZED,atts);
					th.characters(de.getVocalization().toCharArray(),0,de.getVocalization().length());
					th.endElement("","",VOCALIZED);
					//add morphological category
					th.startElement("","",MORPHOLOGICAL_CATEGORY,atts);
					th.characters(de.getMorphology().toCharArray(),0,de.getMorphology().length());
					th.endElement("","",MORPHOLOGICAL_CATEGORY);
					//add glosses
					th.startElement("","",GLOSSES,atts);
					String[] glo = de.getGlosses();
					for (int i = 0; i < glo.length;i++) {
						th.startElement("","",GLOSS,atts);
						th.characters(glo[i].toCharArray(),0,glo[i].length());
						th.endElement("","",GLOSS);
					}
					th.endElement("","",GLOSSES);
					//add grammatical categories
					th.startElement("","",GRAMMATICAL_CATEGORIES,atts);
					String[] pos = de.getPOS();
					for (int i = 0; i < pos.length;i++) {
						th.startElement("","",GRAMMATICAL_CATEGORY,atts);
						String[] parts = pos[i].split("/");
						parts[0] = AraMorph.arabizeWord(parts[0]);
						String arabizedPOS = parts[0];
						if (parts.length > 1)
							arabizedPOS += "/"+parts[1];
						th.characters(arabizedPOS.toCharArray(),0,arabizedPOS.length());
						th.endElement("","",GRAMMATICAL_CATEGORY);
					}
					th.endElement("","",GRAMMATICAL_CATEGORIES);
					//add root
					if (!root.equals("")) {
						th.startElement("","",ROOT,atts);
						th.characters(root.toCharArray(),0,root.length());
						th.endElement("","",ROOT);
					}
					//add lemma
					if (!lemmaID.equals("")) {
						th.startElement("","",LEMMA_ID,atts);
						th.characters(lemmaID.toCharArray(),0,lemmaID.length());
						th.endElement("","",LEMMA_ID);
					}
					else if (!root.equals(""))
						System.err.println("Found an element without lemma at line: "+IN.getLineNumber());

					//append dictionary entry to xml file
					th.endElement("","",ENTRY);
					forms++;
				}
			}
			IN.close();
			//finalize xml file
			th.endElement("","","dictionary");
			th.endDocument();
			//write status message
			System.out.println();
			//if (!"".equals(lemmaID)) System.out.print(lemmas.size() + " lemmas and ");
			//System.out.println(xmlRoot.getChildNodes().getLength() + " entries totalizing " + forms + " forms");
		} catch (TransformerConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			throw new RuntimeException("Can not open : " + name);
		}
	}

	/**
	 * This function translates the given gloss according
	 * to the translate map. The characters of the gloss
	 * will be traversed one by one and replacted according
	 * to the translation map.
	 * @param gloss the gloss to be translated.
	 */
	private String translate(String gloss) {
		char[] translatedGloss = new char[gloss.length() * 2];
		int i=0;
		char[] arGloss = gloss.toCharArray();
		
		for (int x = 0;x < arGloss.length;x++) {
			Object result = translationMap.get(new Character(arGloss[x]));
			if (result == null)
				translatedGloss[i++] = arGloss[x];
			else if (result instanceof Character) {
				Character temp1 = (Character)result;
				translatedGloss[i++] = temp1.charValue();
			} else /*instance of String*/{
				String temp1 = (String)result;
				char temp2[] = temp1.toCharArray();
				for (int j = 0; j < temp2.length;j++)
					translatedGloss[i++]=temp2[j];
			}
		}
		return new String(translatedGloss,0,i);
	}

	/**
	 * This function transforms a compatability table from text format to XML file.
	 * @param firstPart name of the first part of each entry. Either "prefix", "suffix" or "stem"
	 * @param secondPart name of the first part of each entry. Either "prefix", "suffix" or "stem"
	 * @param name name of the compatability table. This name is used as the output file name.
	 * @param is InputStream for the text file to be converted.
	 * @throws RuntimeException if any IO Exception happened during reading out of the InputStream
	 */
	private static void transformCompatibilityTable(String firstPart,String secondPart,String name, InputStream is) throws RuntimeException {
		System.out.print("Transforming compatibility table : " + name + " ");
		try {
			//prepare output file
			StreamResult out = new StreamResult(name+".xml");
			//prpare XML parser
			SAXTransformerFactory tf = (SAXTransformerFactory) SAXTransformerFactory.newInstance();
			TransformerHandler th = tf.newTransformerHandler();
			Transformer serializer = th.getTransformer();
			serializer.setOutputProperty(OutputKeys.ENCODING,XML_ENCODING);
			//pretty XML output
		    serializer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
		    serializer.setOutputProperty(OutputKeys.INDENT,"yes");
		    //attach output file to XML
		    th.setResult(out);
		    //initialize attributes
		    AttributesImpl atts = new AttributesImpl();

		    //start writing the output file
			// Create an XML Document
			th.startDocument();
			th.startElement("","","compatability-table",atts);
			
			LineNumberReader IN = new LineNumberReader(new InputStreamReader(is,"ISO8859_1"));
			String line = null;
			while ((line = IN.readLine()) != null) {
				if ((IN.getLineNumber() % 1000) == 1) System.out.print(".");
				if (!line.startsWith(";")) { //Ignore comments
					line = line.trim();
					line = line.replaceAll("\\s+", " ");
					String[] parts = line.split(" ");
					th.startElement("","",ENTRY,atts);
					th.startElement("","",firstPart,atts);
					th.characters(parts[0].toCharArray(),0,parts[0].length());
					th.endElement("","",firstPart);
					th.startElement("","",secondPart,atts);
					th.characters(parts[1].toCharArray(),0,parts[1].length());
					th.endElement("","",secondPart);
					th.endElement("","",ENTRY);
					//set.add(line);
				} else {
					String comment = line.substring(1).trim();
					th.comment(comment.toCharArray(),0,comment.length());
				}

			}
			IN.close();
			System.out.println();
			//System.out.println(set.size() + " entries");
			th.endElement("","","compatability-table");
			th.endDocument();
		}
		catch (IOException e) {
			throw new RuntimeException("Can not open : " + name);
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (TransformerConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
