package gpl.pierrick.brihaye.aramorph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This classes holds data for one dictionary.
 * @author Ahmed Saad, 2005
 */
public class Dictionary {
	/**A map that stores all dictionary entries. Each unvocalized entry maps to many dictionary entries*/
	private Map entries = new HashMap();
	
	/**Default contructor*/
	public Dictionary() {}
	
	/**Adds a dictionary entry to this dictionary*/
	public void addDictionaryEntry(DictionaryEntry de) {
		Object o = entries.get(de.getEntry());
		if (o == null) {
			o = new ArrayList();
			entries.put(de.getEntry(),o);
		}
		Collection c = (Collection)o;
		c.add(de);
	}
	
	/**Checks whether this dictionary contains entries with the given unvocalized*/
	public boolean contains(String entry) {
		return entries.containsKey(entry);
	}
	
	/**Returns an iterator for all dictionary entries with the given unvocalized*/
	public Iterator getIterator(String entry) {
		if (!entries.containsKey(entry))
			return null;
		return ((Collection) entries.get(entry)).iterator();
	}
}

/**
 * Loads a dictionary from an XML file.
 * @author Ahmed Saad
 */
class XMLDictionaryLoader extends DefaultHandler {
	/**Tells whether this XML file is hierarchical of list of entries.
	 * It is known from the first tag after the root tag.
	 * If it is &lt;entry&gt; tag, the XML file is not hierarchical.
	 * If it is &lt;root&gt; tag, the XML file is hierarchical.*/
	private boolean isHierarchicalXML;
	/**Is the first tag after the root has been reached or not.
	 * Used to correctly determine the type of XML file (hierarchical or flat)*/
	private boolean isFirstTag = true;
	/**The dictionary to be loaded*/
	private Dictionary dictionary;
	/**The unvocalized of the dictionary entry*/
	private String unvocalized;
	/**The vocalized of the dictionary entry*/
	private String vocalized;
	/**The morphologicalCategory of the dictionary entry*/
	private String morphologicalCateogory;
	/**The glosses of the dictionary entry*/
	private Collection glosses;
	/**The grammaticalCategories of the dictionary entry*/
	private Collection grammaticalCategories;
	/**The lemmaID of the dictionary entry*/
	private String lemmaID;
	/**The root of the dictionary entry*/
	private String root;
	
	/**The current location in the XML file*/
	private Integer location;
	/**A map that maps each tag of XML file to its corresponding location*/
	private static Map locations = getLocationsMap();

	/**
	 * Creates a new DictionaryLoader that will load the passed dictionary.
	 * @param d the dictionary to be loaded. This must have been constructed and may not be null.
	 */
	public XMLDictionaryLoader(Dictionary d) {
		this.dictionary = d;
	}

	public void characters(char[] ch, int start, int length) throws SAXException {
		int loc = location.intValue();
		String value = new String(ch,start,length).trim();
		if (value.length() == 0)
			return;
		switch (loc) {
		case 1: unvocalized = value; break;
		case 2: vocalized = value; break;
		case 3: morphologicalCateogory = value; break;
		case 4: glosses.add(value); break;
		case 5: grammaticalCategories.add(value); break;
		case 6: lemmaID = value; break;
		case 7: root = value; break;
		}
		
	}

	public void endElement(String uri, String localName, String qName) throws SAXException {
		int location = ((Integer)locations.get(qName)).intValue();
		if (location == 0) {
			//entry
			String[] arGlosses = new String[glosses.size()];
			String[] arGrammaticalCategories = new String[grammaticalCategories.size()];
			DictionaryEntry de = new DictionaryEntry(unvocalized, lemmaID,
					vocalized, morphologicalCateogory, (String[]) glosses.toArray(arGlosses),
					(String[]) grammaticalCategories.toArray(arGrammaticalCategories),root);
			dictionary.addDictionaryEntry(de);
		} else if (location == 6 && isHierarchicalXML)
			lemmaID = "";
		else if (location == 7 && isHierarchicalXML) {
			root = "";
			lemmaID = "";
		}
			
	}

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		location = (Integer)locations.get(qName);
		int newLocation = location.intValue();
		if (newLocation == 0 && isFirstTag) {
			isHierarchicalXML = false;
			isFirstTag = false;
		} else if (newLocation == 7 && isFirstTag) {
			isHierarchicalXML = true;
			isFirstTag = false;
		}
		if (newLocation == 0) {
			//new entry -> reset the dictionary entry
			unvocalized = "";
			vocalized = "";
			morphologicalCateogory = "";
			glosses = new ArrayList();
			grammaticalCategories = new ArrayList();
			if (!isHierarchicalXML) {
				lemmaID = "";
				root = "";
			}
		} else if (newLocation == 6 && isHierarchicalXML) {
			//a lemma ID that is not inside <entry> tage.
			//we get the lemma from the attribute rather than the text #PCDATA
			lemmaID = attributes.getValue(0);
			//System.err.println(lemmaID);
		} else if (newLocation == 7 && isHierarchicalXML) {
			//a root that is not inside <entry> tage.
			//we get the root from the attribute rather than the text #PCDATA
			root = attributes.getValue(0);
		}
	}

	/**
	 * Creates the tag to location map.
	 * @return a map that maps each tag name to location.
	 */
	private static Map getLocationsMap() {
		Map locations = new HashMap();
		locations.put("dictionary",new Integer(-1));
		locations.put("entry",new Integer(0));
		locations.put("unvocalized",new Integer(1));
		locations.put("vocalized",new Integer(2));
		locations.put("morphological-category",new Integer(3));
		locations.put("glosses",new Integer(4));
		locations.put("gloss",new Integer(4));
		locations.put("grammatical-categories",new Integer(5));
		locations.put("grammatical-category",new Integer(5));
		locations.put("lemma",new Integer(6));
		locations.put("lemmaID",new Integer(6));
		locations.put("root",new Integer(7));
		return locations;
	}
}