/*
Copyright (C) 2003  Pierrick Brihaye
pierrick.brihaye@wanadoo.fr
 
Original Perl code :
Portions (c) 2002 QAMUS LLC (www.qamus.org), 
(c) 2002 Trustees of the University of Pennsylvania 
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the
Free Software Foundation, Inc.
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
or connect to:
http://www.fsf.org/copyleft/gpl.html
*/

package gpl.pierrick.brihaye.aramorph;

import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;

/** An in-memory dictionary of prefixes, stems, suffixes and combinations fed with
 * resources avalaible in the classpath.
 * TODO : use a Lucene index ;-) or any other fast-access resources.
 *@author Pierrick Brihaye, 2003
 */
class SAXXMLDictionaryHandler implements DictionaryHandler{
	/** The unique instance of this handler. */
	private static SAXXMLDictionaryHandler handler = null;
	/** Dictionary of prefixes */
	private Dictionary prefixes;
	/** Dictionary of stems */
	private Dictionary stems;
	/** Dictionary of suffixes */
	private Dictionary suffixes;
	/** Compatibility table for prefixes-stems combinations.*/
	private CompatibilityTable tableAB;
	/** Compatibility table for prefixes-suffixes combinations.*/
	private CompatibilityTable tableAC;
	/** Compatibility table for stems-suffixes combinations.*/
	private CompatibilityTable tableBC;
	
	/** Private constructor to avoid multiple instanciations.*/
	private SAXXMLDictionaryHandler() {
		System.out.println("Initializing in-memory dictionary handler...");
		// load 3 lexicons
		prefixes = loadDictionary(this.getClass().getResourceAsStream("dictionaries/dictPrefixes.xml"));
		stems = loadDictionary(this.getClass().getResourceAsStream("dictionaries/dictStems.xml"));
		suffixes = loadDictionary(this.getClass().getResourceAsStream("dictionaries/dictSuffixes.xml"));
		//load 3 compatibility tables
		tableAB = loadCompatibilityTable(this.getClass().getResourceAsStream("dictionaries/tableAB.xml"));
		tableAC = loadCompatibilityTable(this.getClass().getResourceAsStream("dictionaries/tableAC.xml"));
		tableBC = loadCompatibilityTable(this.getClass().getResourceAsStream("dictionaries/tableBC.xml"));
		
		handler = this;
		System.out.println("... done.");
	}

	/** Returns a unique instance of the handler.
	 * @return The instance
	 */
	protected static synchronized SAXXMLDictionaryHandler getHandler() {
		if (handler == null) 
			handler = new SAXXMLDictionaryHandler();
		return handler;
	}
	
	/** Whether or not the prefix is in the dictionary.
	 * @param translitered The prefix
	 * @return The result
	 */
	public boolean hasPrefix(String translitered) {
		return prefixes.contains(translitered);
	}
	
	/** Returns an iterator on the solutions for the given prefix.
	 * @param translitered The prefix
	 * @return The iterator
	 */
	public Iterator getPrefixIterator(String translitered) {
		return prefixes.getIterator(translitered);
	}
	
	/** Whether or not the stem is in the dictionary.
	 * @param translitered The stem
	 * @return The result
	 */
	public boolean hasStem(String translitered) {
		return stems.contains(translitered);
	}
	
	/** Returns an iterator on the solutions for the given stem.
	 * @param translitered The stem
	 * @return The iterator
	 */
	public Iterator getStemIterator(String translitered) {
		return stems.getIterator(translitered);
	}
	
	/** Whether or not the suffix is in the dictionary.
	 * @param translitered The suffix
	 * @return The result
	 */
	public boolean hasSuffix(String translitered) {
		return suffixes.contains(translitered);
	}
	
	/** Returns an iterator on the solutions for the given suffix.
	 * @param translitered The suffix
	 * @return The iterator
	 */
	public Iterator getSuffixIterator(String translitered) {
		return suffixes.getIterator(translitered);
	}
	
	/** Whether or not the prefix/stem combination is possible.
	 * @param AB The prefix and stem combination.
	 * @return The result
	 */
	public boolean hasAB(String A, String B) {
		return tableAB.contains(A,B);
	}
	
	/** Whether or not the prefix/suffix combination is possible.
	 * @param AC The prefix and suffix combination.
	 * @return The result
	 */
	public boolean hasAC(String A, String C) {
		return tableAC.contains(A,C);
	}
	
	/** Whether or not the stem/suffix combination is possible.
	 * @param BC The stem and suffix combination.
	 * @return The result
	 */
	public boolean hasBC(String B, String C) {
		return tableBC.contains(B,C);
	}
	
	/**
	 * Loads a dictionary from an XML file.
	 * The file is already open and its input stream is passed.
	 * @param is the input stream for the XML file.
	 * @return the dictionary that has been loaded.
	 */
	private Dictionary loadDictionary(InputStream is) {
		Dictionary dict = new Dictionary();
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setValidating(false);
		try {
			SAXParser sax = factory.newSAXParser();
			XMLDictionaryLoader dl = new XMLDictionaryLoader(dict);
			sax.parse(is,dl);
		} catch (ParserConfigurationException pce) {
			System.err.println("Could not create that parser.");
			System.err.println(pce.getMessage());
		} catch (SAXException se) {
			System.err.println("Problem with the SAX parser");
			System.err.println(se.getMessage());
		} catch (IOException ioe) {
			System.err.println("Error reading file.");
			System.err.println(ioe.getMessage());
		}
		return dict;
	}

	/**
	 * Loads a compatibility table from an XML file.
	 * The XML file is already open and its input stream is passed.
	 * @param is the input stream for the XML file.
	 * @return the Compatibility table that has been loaded.
	 */
	private CompatibilityTable loadCompatibilityTable(InputStream is) {
		CompatibilityTable table = new CompatibilityTable();
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setValidating(false);
		try {
			SAXParser sax = factory.newSAXParser();
			XMLCompatibilityLoader cl = new XMLCompatibilityLoader(table);
			sax.parse(is,cl);
		} catch (ParserConfigurationException pce) {
			System.err.println("Could not create that parser.");
			System.err.println(pce.getMessage());
		} catch (SAXException se) {
			System.err.println("Problem with the SAX parser");
			System.err.println(se.getMessage());
		} catch (IOException ioe) {
			System.err.println("Error reading file.");
			System.err.println(ioe.getMessage());
		}
		return table;
	}
	
}


